//
//  l1f10085.c
//  Graphic LCD module "EPSON L1F10085" I/O library
//  Copyright(C)2012 Toyohiko Togashi tog001@nifty.com
//
//
//  This program is free software; you can redistribute it and/or modify it under the terms of the
//  GNU General Public License as published by the Free Software Foundation; either version 3
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along with this program.
//  If not, see <http://www.gnu.org/licenses/>
//
//
//    Update history
//    ---------- ----- -------------------------------------------
//    2012.02.20 v0.0  First cording
//    2012.04.18 v0.1  Bug fix, '\r' function
//

#include <avr/io.h>						// WinAVR library
#include <avr/pgmspace.h>				// Flash access
#include <avr/interrupt.h>				// 
#include <util/delay.h>					// 

#include "l1f10085.h"					//

#define ABS(a)		(a < 0 ? -a : a)
#define OFFSET_X	0
#define OFFSET_Y	2

//
// 	Port I/O
//
void lcd_sendCmdData(unsigned char dc, unsigned char data) {
	unsigned char	i;
//	unsigned char	s;

//	s = SREG;
//	cli();

	LCD_PIN_SCK_OFF();
	LCD_PIN_CS_ON();

	if (dc) {
		LCD_PIN_SDATA_ON();
	} else {
		LCD_PIN_SDATA_OFF();
	}
	LCD_PIN_SCK_ON();
	LCD_PIN_SCK_OFF();
	for(i = 0; i < 8; i++) {
		if (data & 0x80) {
			LCD_PIN_SDATA_ON();
		} else {
			LCD_PIN_SDATA_OFF();
		}
		LCD_PIN_SCK_ON();
		LCD_PIN_SCK_OFF();
		data = data << 1;
	}

	LCD_PIN_CS_OFF();

//	if (s & _BV(SREG_I)) {				// Resume global interrupt
//		sei();
//	}

	return;
}

//
//	Open
//
#define LCD_DC_WAIT			2
#define LCD_DC_END			99

const prog_uchar lcd_init_cmd[] = {		// Flash storage
	LCD_DC_COMMAND, LCD_CMD_COMSCN,		// Common scan direction
		LCD_DC_DATA, 0b00000001,		//		b0:pin1->68reverse b1:pin69->132reverse  <L1F10085>
	LCD_DC_COMMAND, LCD_CMD_OSCON,		// Internal oscillation on
	LCD_DC_COMMAND, LCD_CMD_SLPOUT,		// Sleep out
	LCD_DC_COMMAND, LCD_CMD_PWRCTR,		// Power control
		LCD_DC_DATA, 0b00001111,		//		b0:Reference b1:DriverCircuit b2:2nd-booster b3:prim-booster
	LCD_DC_COMMAND, LCD_CMD_DATCTL,		// Data scan direction, etc.
		LCD_DC_DATA, 0b00000010,		//		b0:Page b1:Column b2:Scan address direction inverse   <L1F10085>
		LCD_DC_DATA, 0b00000001,		//		b0:BGR  <L1F10085>
		LCD_DC_DATA, 0b00000001,		//		b0:256colors b1:4096colorsTypeA(2pixel:2byte) b1:4096colorsTypeB(1pixel:2byte)
	LCD_DC_COMMAND, LCD_CMD_RGBSET8,	// 256-color position set
		LCD_DC_DATA, 0,					//		Red 000
		LCD_DC_DATA, 2,					//		    001
		LCD_DC_DATA, 4,					//		    010
		LCD_DC_DATA, 6,					//		    011
		LCD_DC_DATA, 8,					//		    100
		LCD_DC_DATA, 10,				//		    101
		LCD_DC_DATA, 12,				//		    110
		LCD_DC_DATA, 15,				//		    111
		LCD_DC_DATA, 0,					//		Green 000
		LCD_DC_DATA, 2,					//		      001
		LCD_DC_DATA, 4,					//		      010
		LCD_DC_DATA, 6,					//		      011
		LCD_DC_DATA, 8,					//		      100
		LCD_DC_DATA, 10,				//		      101
		LCD_DC_DATA, 12,				//		      110
		LCD_DC_DATA, 15,				//		      111
		LCD_DC_DATA, 0,					//		Blue 00
		LCD_DC_DATA, 6,					//		     01
		LCD_DC_DATA, 9,					//		     10
		LCD_DC_DATA, 15,				//		     11
	LCD_DC_COMMAND, LCD_CMD_VOLCTR,		// Electronic volume control
		LCD_DC_DATA, 40,				//		0..63
		LCD_DC_DATA, 3,					//		0..7
	LCD_DC_WAIT, 100,					// 
	LCD_DC_COMMAND, LCD_CMD_DISON,		// Display on
	LCD_DC_END, 0						//
};

void lcd_start(void) {
#ifdef LCD_PIN_RESET_INIT
	LCD_PIN_RESET_INIT();
#endif
	LCD_PIN_SDATA_INIT();
	LCD_PIN_CS_INIT();
	LCD_PIN_SCK_INIT();
	return;
}

void lcd_init(void) {
	int i;

	// Reset sequence
#ifdef LCD_PIN_RESET_ON
	LCD_PIN_RESET_ON();
	_delay_ms(200);
#endif
#ifdef LCD_PIN_RESET_OFF
	LCD_PIN_RESET_OFF();
	_delay_ms(200);
#endif


	// Init code
	for(i = 0; ; i += 2) {
		unsigned char	dc;
		unsigned char	data;

		dc   = pgm_read_byte(&lcd_init_cmd[i]);
		data = pgm_read_byte(&lcd_init_cmd[i+1]);
		if (dc >= LCD_DC_END) {
			break;
		} else if (dc == LCD_DC_WAIT) {
//			_delay_ms(data);			// Code is big
			_delay_ms(100);
		} else {
			lcd_sendCmdData(dc, data);
		}
	}

	// Screen clear
	lcd_draw_rect(0, 0, LCD_SCREEN_WIDTH, LCD_SCREEN_HEIGHT, LCD_COLOR_BLACK);

	return;
}

//
//	Close
//
void lcd_stop(void) {
	LCD_PIN_SCK_TERM();
	LCD_PIN_CS_TERM();
	LCD_PIN_SDATA_TERM();
#ifdef LCD_PIN_RESET_TERM
	LCD_PIN_RESET_TERM();
#endif
	return;
}

//
//	Rectangle
//
void lcd_draw_rect(
	unsigned char	x,					// Upper right position
	unsigned char	y,					//
	unsigned char	w,					// width
	unsigned char	h,					// height
	unsigned char	color				//
) {
	short	i;
	short	m;

	x += OFFSET_X;
	y += OFFSET_Y;

	color = ~color;
	lcd_sendCmdData(0, LCD_CMD_PASET);	lcd_sendCmdData(1, y);	lcd_sendCmdData(1, y+h-1);
	lcd_sendCmdData(0, LCD_CMD_CASET);	lcd_sendCmdData(1, x);	lcd_sendCmdData(1, x+w-1);
	lcd_sendCmdData(0, LCD_CMD_RAMWR);
	m = (short)h * (short)w;
	for(i = 0; i < m; i++) {
		lcd_sendCmdData(1, color);
	}
	return;
}

//
//	Line
//
void lcd_draw_line(
	unsigned char	x1,					// Start point
	unsigned char	y1,					//
	unsigned char	x2,					// End point
	unsigned char	y2,					//
	unsigned char	s,					// Line size
	unsigned char	color				//
) {
	short			w;
	short			h;
	short			i;

	w = (short)x2 - (short)x1 + 1;
	h = (short)y2 - (short)y1 + 1;
//	if (w != 0) {
//		w++;
//	}
//	if (h != 0) {
//		h++;
//	}

	if (ABS(w) < ABS(h)) {				// 45..90deg
		if (h < 0) {
			for(i = 0; i > h; i--) {
				unsigned char	x,y;

				x = x1 + (i * w / h);
				y = y1 + i;
				lcd_draw_rect(x, y, s, s, color);
			}
		} else {
			for(i = 0; i < h; i++) {
				unsigned char	x,y;

				x = x1 + (i * w / h);
				y = y1 + i;
				lcd_draw_rect(x, y, s, s, color);
			}
		}
	} else {							// 0..45deg
		if (w < 0) {
			for(i = 0; i > w; i--) {
				unsigned char	x,y;

				x = x1 + i;
				y = y1 + (i * h / w);
				lcd_draw_rect(x, y, s, s, color);
			}
		} else {
			for(i = 0; i < w; i++) {
				unsigned char	x,y;

				x = x1 + i;
				y = y1 + (i * h / w);
				lcd_draw_rect(x, y, s, s, color);
			}
		}
	}

	return;
}

//
//	Character font
//
const prog_uchar	lcd_font5x7[][8] = {
	{0b00000,	// " "
	 0b00000,
	 0b00000,
	 0b00000,
	 0b00000,
	 0b00000,
	 0b00000,
	 0},
	{0b00100,	// !
	 0b00100,
	 0b00100,
	 0b00100,
	 0b00100,
	 0b00000,
	 0b00100,
	 0},
	{0b01010,	// "
	 0b01010,
	 0b00000,
	 0b00000,
	 0b00000,
	 0b00000,
	 0b00000,
	 0},
	{0b01010,	// #
	 0b01010,
	 0b11111,
	 0b01010,
	 0b01010,
	 0b11111,
	 0b01010,
	 0},
	{0b00100,	// $
	 0b01111,
	 0b10100,
	 0b01110,
	 0b00101,
	 0b11111,
	 0b00100,
	 0},
	{0b01001,	// %
	 0b10101,
	 0b01010,
	 0b00100,
	 0b01010,
	 0b10101,
	 0b10010,
	 0},
	{0b01100,	// &
	 0b10010,
	 0b10100,
	 0b01000,
	 0b10101,
	 0b10010,
	 0b01101,
	 0},
	{0b00100,	// '
	 0b00100,
	 0b00000,
	 0b00000,
	 0b00000,
	 0b00000,
	 0b00000,
	 0},
	{0b00010,	// (
	 0b00100,
	 0b01000,
	 0b01000,
	 0b01000,
	 0b00100,
	 0b00010,
	 0},
	{0b01000,	// )
	 0b00100,
	 0b00010,
	 0b00010,
	 0b00010,
	 0b00100,
	 0b01000,
	 0},
	{0b00000,	// *
	 0b00100,
	 0b10101,
	 0b01110,
	 0b10101,
	 0b00100,
	 0b00000,
	 0},
	{0b00000,	// +
	 0b00100,
	 0b00100,
	 0b11111,
	 0b00100,
	 0b00100,
	 0b00000,
	 0},
	{0b00000,	// ,
	 0b00000,
	 0b00000,
	 0b00000,
	 0b01100,
	 0b00100,
	 0b01000,
	 0},
	{0b00000,	// -
	 0b00000,
	 0b00000,
	 0b11111,
	 0b00000,
	 0b00000,
	 0b00000,
	 0},
	{0b00000,	// .
	 0b00000,
	 0b00000,
	 0b00000,
	 0b00000,
	 0b01100,
	 0b01100,
	 0},
	{0b00000,	// /
	 0b00001,
	 0b00010,
	 0b00100,
	 0b01000,
	 0b10000,
	 0b00000,
	 0},
	{0b01110,	// 0
	 0b10001,
	 0b10001,
	 0b10001,
	 0b10001,
	 0b10001,
	 0b01110,
	 0},
	{0b00100,	// 1
	 0b01100,
	 0b00100,
	 0b00100,
	 0b00100,
	 0b00100,
	 0b00100,
	 0},
	{0b01110,	// 2
	 0b10001,
	 0b00001,
	 0b00010,
	 0b00100,
	 0b01000,
	 0b11111,
	 0},
	{0b01110,	// 3
	 0b10001,
	 0b00001,
	 0b00110,
	 0b00001,
	 0b10001,
	 0b01110,
	 0},
	{0b10010,	// 4
	 0b10010,
	 0b10010,
	 0b11111,
	 0b00010,
	 0b00010,
	 0b00010,
	 0},
	{0b11111,	// 5
	 0b10000,
	 0b11110,
	 0b00001,
	 0b00001,
	 0b10001,
	 0b01110,
	 0},
	{0b01110,	// 6
	 0b10001,
	 0b10000,
	 0b11110,
	 0b10001,
	 0b10001,
	 0b01110,
	 0},
	{0b11111,	// 7
	 0b00001,
	 0b00010,
	 0b00100,
	 0b00100,
	 0b00100,
	 0b00100,
	 0},
	{0b01110,	// 8
	 0b10001,
	 0b10001,
	 0b01110,
	 0b10001,
	 0b10001,
	 0b01110,
	 0},
	{0b01110,	// 9
	 0b10001,
	 0b10001,
	 0b01111,
	 0b00001,
	 0b10001,
	 0b01110,
	 0},
	{0b00000,	// :
	 0b01100,
	 0b01100,
	 0b00000,
	 0b01100,
	 0b01100,
	 0b00000,
	 0},
	{0b00000,	// ;
	 0b01100,
	 0b01100,
	 0b00000,
	 0b01100,
	 0b00100,
	 0b01000,
	 0},
	{0b00010,	// <
	 0b00100,
	 0b01000,
	 0b10000,
	 0b01000,
	 0b00100,
	 0b00010,
	 0},
	{0b00000,	// =
	 0b00000,
	 0b11111,
	 0b00000,
	 0b11111,
	 0b00000,
	 0b00000,
	 0},
	{0b01000,	// >
	 0b00100,
	 0b00010,
	 0b00001,
	 0b00010,
	 0b00100,
	 0b01000,
	 0},
	{0b01110,	// ?
	 0b10001,
	 0b00001,
	 0b00010,
	 0b00100,
	 0b00000,
	 0b00100,
	 0},
	{0b01110,	// @
	 0b10001,
	 0b10111,
	 0b10101,
	 0b10111,
	 0b10000,
	 0b01110,
	 0},
	{0b00100,	// A
	 0b01010,
	 0b10001,
	 0b10001,
	 0b11111,
	 0b10001,
	 0b10001,
	 0},
	{0b11110,	// B
	 0b10001,
	 0b10001,
	 0b11110,
	 0b10001,
	 0b10001,
	 0b11110,
	 0},
	{0b01110,	// C
	 0b10001,
	 0b10000,
	 0b10000,
	 0b10000,
	 0b10001,
	 0b01110,
	 0},
	{0b11110,	// D
	 0b10001,
	 0b10001,
	 0b10001,
	 0b10001,
	 0b10001,
	 0b11110,
	 0},
	{0b11111,	// E
	 0b10000,
	 0b10000,
	 0b11111,
	 0b10000,
	 0b10000,
	 0b11111,
	 0},
	{0b11111,	// F
	 0b10000,
	 0b10000,
	 0b11111,
	 0b10000,
	 0b10000,
	 0b10000,
	 0},
	{0b01110,	// G
	 0b10001,
	 0b10000,
	 0b10000,
	 0b10011,
	 0b10001,
	 0b01110,
	 0},
	{0b10001,	// H
	 0b10001,
	 0b10001,
	 0b11111,
	 0b10001,
	 0b10001,
	 0b10001,
	 0},
	{0b01110,	// I
	 0b00100,
	 0b00100,
	 0b00100,
	 0b00100,
	 0b00100,
	 0b01110,
	 0},
	{0b00010,	// J
	 0b00010,
	 0b00010,
	 0b00010,
	 0b00010,
	 0b10010,
	 0b01100,
	 0},
	{0b10001,	// K
	 0b10010,
	 0b10100,
	 0b11000,
	 0b10100,
	 0b10010,
	 0b10001,
	 0},
	{0b10000,	// L
	 0b10000,
	 0b10000,
	 0b10000,
	 0b10000,
	 0b10000,
	 0b11111,
	 0},
	{0b10001,	// M
	 0b11011,
	 0b10101,
	 0b10001,
	 0b10001,
	 0b10001,
	 0b10001,
	 0},
	{0b10001,	// N
	 0b10001,
	 0b11001,
	 0b10101,
	 0b10011,
	 0b10001,
	 0b10001,
	 0},
	{0b01110,	// O
	 0b10001,
	 0b10001,
	 0b10001,
	 0b10001,
	 0b10001,
	 0b01110,
	 0},
	{0b11110,	// P
	 0b10001,
	 0b10001,
	 0b11110,
	 0b10000,
	 0b10000,
	 0b10000,
	 0},
	{0b01110,	// Q
	 0b10001,
	 0b10001,
	 0b10001,
	 0b10101,
	 0b10010,
	 0b01101,
	 0},
	{0b11110,	// R
	 0b10001,
	 0b10001,
	 0b11110,
	 0b10100,
	 0b10010,
	 0b10001,
	 0},
	{0b01110,	// S
	 0b10001,
	 0b10000,
	 0b01110,
	 0b00001,
	 0b10001,
	 0b01110,
	 0},
	{0b11111,	// T
	 0b00100,
	 0b00100,
	 0b00100,
	 0b00100,
	 0b00100,
	 0b00100,
	 0},
	{0b10001,	// U
	 0b10001,
	 0b10001,
	 0b10001,
	 0b10001,
	 0b10001,
	 0b01110,
	 0},
	{0b10001,	// V
	 0b10001,
	 0b10001,
	 0b10001,
	 0b10001,
	 0b01010,
	 0b00100,
	 0},
	{0b10001,	// W
	 0b10001,
	 0b10001,
	 0b10101,
	 0b10101,
	 0b10101,
	 0b01010,
	 0},
	{0b10001,	// X
	 0b10001,
	 0b01010,
	 0b00100,
	 0b01010,
	 0b10001,
	 0b10001,
	 0},
	{0b10001,	// Y
	 0b01010,
	 0b00100,
	 0b00100,
	 0b00100,
	 0b00100,
	 0b00100,
	 0},
	{0b11111,	// Z
	 0b00001,
	 0b00010,
	 0b00100,
	 0b01000,
	 0b10000,
	 0b11111,
	 0},
	{0b01110,	// [
	 0b01000,
	 0b01000,
	 0b01000,
	 0b01000,
	 0b01000,
	 0b01110,
	 0},
	{0b00000,	// 
	 0b10000,
	 0b01000,
	 0b00100,
	 0b00010,
	 0b00001,
	 0b00000,
	 0},
	{0b01110,	// ]
	 0b00010,
	 0b00010,
	 0b00010,
	 0b00010,
	 0b00010,
	 0b01110,
	 0},
	{0b00100,	// ^
	 0b01010,
	 0b00000,
	 0b00000,
	 0b00000,
	 0b00000,
	 0b00000,
	 0},
	{0b00000,	// _
	 0b00000,
	 0b00000,
	 0b00000,
	 0b00000,
	 0b00000,
	 0b11111,
	 0},
	{0b01000,	// `
	 0b00100,
	 0b00000,
	 0b00000,
	 0b00000,
	 0b00000,
	 0b00000,
	 0},
	{0b00000,	// a
	 0b00000,
	 0b01110,
	 0b00001,
	 0b01111,
	 0b10011,
	 0b01101,
	 0},
	{0b10000,	// b
	 0b10000,
	 0b10000,
	 0b11110,
	 0b10001,
	 0b10001,
	 0b11110,
	 0},
	{0b00000,	// c
	 0b00000,
	 0b00000,
	 0b01111,
	 0b10000,
	 0b10000,
	 0b01111,
	 0},
	{0b00001,	// d
	 0b00001,
	 0b00001,
	 0b01111,
	 0b10001,
	 0b10001,
	 0b01111,
	 0},
	{0b00000,	// e
	 0b00000,
	 0b01110,
	 0b10001,
	 0b11111,
	 0b10000,
	 0b01110,
	 0},
	{0b00011,	// f
	 0b00100,
	 0b00100,
	 0b01110,
	 0b00100,
	 0b00100,
	 0b00100,
	 0},
	{0b00000,	// g
	 0b00000,
	 0b01110,
	 0b10001,
	 0b01111,
	 0b00001,
	 0b01110,
	 0},
	{0b10000,	// h
	 0b10000,
	 0b10000,
	 0b11110,
	 0b10001,
	 0b10001,
	 0b10001,
	 0},
	{0b00100,	// i
	 0b00000,
	 0b00100,
	 0b00100,
	 0b00100,
	 0b00100,
	 0b00100,
	 0},
	{0b00100,	// j
	 0b00000,
	 0b00100,
	 0b00100,
	 0b00100,
	 0b00100,
	 0b11000,
	 0},
	{0b10000,	// k
	 0b10000,
	 0b10001,
	 0b10010,
	 0b10100,
	 0b11010,
	 0b10001,
	 0},
	{0b00100,	// l
	 0b00100,
	 0b00100,
	 0b00100,
	 0b00100,
	 0b00100,
	 0b00110,
	 0},
	{0b00000,	// m
	 0b00000,
	 0b11010,
	 0b10101,
	 0b10101,
	 0b10001,
	 0b10001,
	 0},
	{0b00000,	// n
	 0b00000,
	 0b11110,
	 0b10001,
	 0b10001,
	 0b10001,
	 0b10001,
	 0},
	{0b00000,	// o
	 0b00000,
	 0b01110,
	 0b10001,
	 0b10001,
	 0b10001,
	 0b01110,
	 0},
	{0b00000,	// p
	 0b00000,
	 0b11110,
	 0b10001,
	 0b11110,
	 0b10000,
	 0b10000,
	 0},
	{0b00000,	// q
	 0b00000,
	 0b01111,
	 0b10001,
	 0b01111,
	 0b00001,
	 0b00001,
	 0},
	{0b00000,	// r
	 0b00000,
	 0b10111,
	 0b11000,
	 0b10000,
	 0b10000,
	 0b10000,
	 0},
	{0b00000,	// s
	 0b00000,
	 0b01110,
	 0b10000,
	 0b01110,
	 0b00001,
	 0b01110,
	 0},
	{0b00000,	// t
	 0b00100,
	 0b01110,
	 0b00100,
	 0b00100,
	 0b00100,
	 0b00010,
	 0},
	{0b00000,	// u
	 0b00000,
	 0b10001,
	 0b10001,
	 0b10001,
	 0b10001,
	 0b01111,
	 0},
	{0b00000,	// v
	 0b00000,
	 0b10001,
	 0b10001,
	 0b10001,
	 0b01010,
	 0b00100,
	 0},
	{0b00000,	// w
	 0b00000,
	 0b10001,
	 0b10001,
	 0b10101,
	 0b10101,
	 0b01010,
	 0},
	{0b00000,	// x
	 0b00000,
	 0b10001,
	 0b01010,
	 0b00100,
	 0b01010,
	 0b10001,
	 0},
	{0b00000,	// y
	 0b00000,
	 0b10001,
	 0b01010,
	 0b00100,
	 0b01000,
	 0b10000,
	 0},
	{0b00000,	// z
	 0b00000,
	 0b11111,
	 0b00010,
	 0b00100,
	 0b01000,
	 0b11111,
	 0},
	{0b00010,	// {
	 0b00100,
	 0b00100,
	 0b01000,
	 0b00100,
	 0b00100,
	 0b00010,
	 0},
	{0b00100,	// |
	 0b00100,
	 0b00100,
	 0b00100,
	 0b00100,
	 0b00100,
	 0b00100,
	 0},
	{0b01000,	// }
	 0b00100,
	 0b00100,
	 0b00010,
	 0b00100,
	 0b00100,
	 0b01000,
	 0},
	{0b00000,	// ~
	 0b00000,
	 0b01000,
	 0b10101,
	 0b00010,
	 0b00000,
	 0b00000,
	 0},
	{0b00100,	// deg
	 0b01010,
	 0b00100,
	 0b00000,
	 0b00000,
	 0b00000,
	 0b00000,
	 0}
};

//
//	Text
//
void lcd_draw_char(unsigned char x, unsigned char y, unsigned char fcolor, unsigned char bcolor, char c) {
	unsigned char	i;
	unsigned char	j;

	x += OFFSET_X;
	y += OFFSET_Y;
	c &= 0x7f;
	c -= ' ';
	lcd_sendCmdData(0, LCD_CMD_PASET);	lcd_sendCmdData(1, y);	lcd_sendCmdData(1, y+LCD_FONT_HEIGHT-1);
	lcd_sendCmdData(0, LCD_CMD_CASET);	lcd_sendCmdData(1, x);	lcd_sendCmdData(1, x+LCD_FONT_WIDTH-1);
	lcd_sendCmdData(0, LCD_CMD_RAMWR);

	for(i = 0; i < LCD_FONT_HEIGHT; i++) {
		unsigned char	b;

		b = pgm_read_byte(&lcd_font5x7[c][i]);
		b <<= (8 - LCD_FONT_WIDTH);
		for(j = 0; j < LCD_FONT_WIDTH; j++) {
			b <<= 1;
			if (b & 0x80) {
				lcd_sendCmdData(1, ~fcolor);
			} else {
				lcd_sendCmdData(1, ~bcolor);
			}
		}
	}
	return;
}

void lcd_draw_str(
	unsigned char	sx,					// Upper right position
	unsigned char	sy,					//
	unsigned char	fcolor,				// Text color
	unsigned char	bcolor,				// Background color
	char			*str				// '\0' terminated text
) {
	unsigned char	x;
	unsigned char	y;
	char			c;

	x  = sx;
	y  = sy;

	while((c = *str++) != '\0') {

		c &= 0b01111111;

		// Visual char
		if (c >= ' ') {
			lcd_draw_char(x, y, fcolor, bcolor, c);
			x += LCD_FONT_WIDTH;

		// CLF
		} else if (c == '\n') {
			x = sx;
			y += LCD_FONT_HEIGHT;

		// CR
		} else if (c == '\r') {
			y += LCD_FONT_HEIGHT;

		}

		// Screen range out
		if (x >= (LCD_SCREEN_WIDTH - LCD_FONT_WIDTH + 1)) {
			x = sx;
			y += LCD_FONT_HEIGHT;
		}
		if (y >= (LCD_SCREEN_HEIGHT - LCD_FONT_HEIGHT + 1)) {
			y = 0;
		}
	}

	return;
}

//
//	Virtual character display interface
//
#ifdef LCD_VCD
unsigned char	lcd_vcd_x1 = 0;
unsigned char	lcd_vcd_y1 = 0;
unsigned char	lcd_vcd_x2 = LCD_SCREEN_WIDTH  - LCD_FONT_WIDTH  + 1;
unsigned char	lcd_vcd_y2 = LCD_SCREEN_HEIGHT - LCD_FONT_HEIGHT + 1;
unsigned char	lcd_vcd_w  = LCD_SCREEN_WIDTH;
unsigned char	lcd_vcd_h  = LCD_SCREEN_HEIGHT;
unsigned char	lcd_curs_x = 0;
unsigned char	lcd_curs_y = 0;
unsigned char	lcd_curs_fcolor = LCD_COLOR_WHITE;
unsigned char	lcd_curs_bcolor = LCD_COLOR_BLACK;

void lcd_setViewPort(unsigned char	x, unsigned char y, unsigned char w, unsigned char h) {
	lcd_vcd_x1 = x;
	lcd_vcd_y1 = y;
	lcd_vcd_x2 = x + w - LCD_FONT_WIDTH  + 1;
	lcd_vcd_y2 = y + h - LCD_FONT_HEIGHT + 1;
	lcd_vcd_w  = w;
	lcd_vcd_h  = h;
	lcd_curs_x = x;
	lcd_curs_y = y;
	return;
}

void lcd_position(unsigned char	column, unsigned char row) {
	lcd_curs_x = column * LCD_FONT_WIDTH  + lcd_vcd_x1;
	lcd_curs_y = row    * LCD_FONT_HEIGHT + lcd_vcd_y1;
	return;
}

void lcd_color(unsigned char fcolor) {
	lcd_curs_fcolor = fcolor;
	return;
}

void lcd_colors(unsigned char fcolor, unsigned char bcolor) {
	lcd_curs_fcolor = fcolor;
	lcd_curs_bcolor = bcolor;
	return;
}

void lcd_prString(char *str) {
	char	c;

	while((c = *str++) != '\0') {

		c &= 0b01111111;

		// Visual char
		if (c >= ' ') {
			lcd_draw_char(lcd_curs_x, lcd_curs_y, lcd_curs_fcolor, lcd_curs_bcolor, c);
			lcd_curs_x += LCD_FONT_WIDTH;

		// LF
		} else if (c == '\n') {
			lcd_curs_x = lcd_vcd_x1;
			lcd_curs_y += LCD_FONT_HEIGHT;

		// CR
		} else if (c == '\r') {
			lcd_curs_x = lcd_vcd_x1;		// ADD 2012.04.18
//			lcd_curs_y += LCD_FONT_HEIGHT;		// DEL 2012.04.18

		// FF
		} else if (c == '\f') {
			lcd_control(LCD_DISP_CLEAR_HOME);

		// VT
		} else if (c == '\v') {
			lcd_control(LCD_CURSOR_HOME);
		}

		// View range out
		if (lcd_curs_x >= lcd_vcd_x2) {
			lcd_curs_x = lcd_vcd_x1;
			lcd_curs_y += LCD_FONT_HEIGHT;
		}
		if (lcd_curs_y >= lcd_vcd_y2) {
			lcd_curs_y = lcd_vcd_y1;
		}
	}

	return;
}

void lcd_control(unsigned char cmd) {
	switch(cmd) {
	case LCD_DISP_CLEAR_HOME:
		lcd_draw_rect(
			lcd_vcd_x1, lcd_vcd_y1,
			lcd_vcd_w, lcd_vcd_h,
			lcd_curs_bcolor);
	case LCD_CURSOR_HOME:
		lcd_curs_x = lcd_vcd_x1;
		lcd_curs_y = lcd_vcd_y1;
		break;
	default:
		break;
	}
	return;
}

#endif	// LCD_VCD


